<?php

  /**
  * Base class for every class used in WApp framework
  */
  abstract class TObject {
    /** @var TCustomConnection */
    private $db = null;
    /**
    * Clone object
    *
    * @return TObject
    */
    public function cloneMe()  {
      return clone $this;
    }
    /**
    *
    *
    * @return void
    */
    protected function cloneProperty($name)  {
      $this->$name = cloneVariable($this->$name);
    }
    protected function evalEvent($code,$arguments) {
      foreach(array_keys($arguments) as $_argName) {
        ${$_argName} = &$arguments[$_argName];
      }
      return eval($code);
    }
    /**
    * Calls function assigned to this event
    *
    * @param string Name of the event to fire
    * @param mixed[] Array of arguments to pass to the function
    * @return mixed Return value of called function or NULL if no function is assigned
    */
    public function fireEvent($eventName, $arguments = array())  {
      if(!isset($this->$eventName)) {      
        throw new EException(E_UNDEFINED_EVENT,array($eventName));
      }
      $eventFunction = $this->$eventName;          
      if($eventFunction) {
        if(substr($eventFunction,0,1) === '>') {
          return $this->evalEvent(substr($eventFunction,1),$arguments);
        }      
        if(!function_exists($eventFunction)) {
          throw new EException(E_EVENT_FUNCTION_NOT_FOUND,array($eventFunction,$eventName));          
        }
        return call_user_func_array($eventFunction,array_merge(array($this),$arguments)); 
      } else {
        return null;
      }
    }
    /**
    * Returns name of class implemented by this object
    *
    * @return string class-name
    */
    public function getClass() {
      return get_class($this);
    }

    /**
    * Returns name of parent of a class implemented by this object
    *
    * @return string class-name
    */
    public function getParentClass() {
      return get_parent_class($this);
    }    
    /**
    * Check if this object is instance of class
    *
    * @param string class-name
    * @return bool
    */
    public function isInstanceOf($classname) {
      return is_subclass_of($this,$classname) || sameStr($classname,$this->getClass());           
    }
    /**
    * Check if this object implements an interface
    *
    * @param string interface-name
    * @return bool
    */
    public function isImplementorOf($classname) {
      $intfs = class_implements($this);
      return isset($intfs[$classname]);          
    }    
    /**
    * Returns array of object's methods
    *
    * @return array
    */
    public function getMethods() {
      return get_class_methods($this);           
    }
    /**
    * Returns array of object's variables
    *
    * @return array
    */
    public function getVars() {
      return get_object_vars($this);           
    }        
    /**
    * Returns array of class defined variables
    *
    * @return array
    */
    public function getClassVars() {
      return get_class_vars($this->getClass());           
    }
    /**
    * Loads all object variables from $obj
    *
    * @param TObject Object 
    * @return void
    */
    public function assign($obj) {
      foreach($obj as $key => $value) {
        $this->$key = $value;
      }
/*
      foreach($obj as $key => $value) {
//        if(isset($obj->$key)) {
          $this->$key = $value;
//        }
      }      
*/
    }
    /**
    * Loads all object variables from $array
    *
    * @param array Array 
    * @return void 
    */
    public function assignArray($array) {
      foreach(array_keys($array) as $key) {
        $this->$key = $array[$key];
      }                 
/*
      foreach(array_keys($array) as $key) {
        if(isset($array[$key])) {
          $this->$key = $array[$key];
        }
      }
*/      
    }
    /**
    * Called when reading an undefined property
    * Do not call directly.
    *
    * @param string Name of the property
    */
    public function __get($prop_name) {
      if(method_exists($this,'_get'.$prop_name)) {
        return $this->{'_get'.$prop_name}();
      } else if(property_exists($this,'_'.$prop_name)) {
        return $this->{'_'.$prop_name};
      } else {
        throw new EException(E_UNDEFINED_PROPERTY,array($this->getClass().'::'.$prop_name));
      } 
    }
    /**
    * Called when writing to an undefined property
    * Do not call directly.
    *
    * @param string Name of the property
    * @param mixed New value to be set    
    */
    public function __set($prop_name, $prop_value) {
      if(method_exists($this,'_set'.$prop_name)) {
        $this->{'_set'.$prop_name}($prop_value);
      } else {
        throw new EException(E_UNDEFINED_PROPERTY,array($this->getClass().'::'.$prop_name));
      }
    }
    /**
    * Set database driver for this object
    *
    * @param TCustomDatabase    
    */
    protected function setDb($db) {
      $this->db = $db;
    }
    /**
    * Get database driver for this object
    *
    * @return TCustomDatabase    
    */
    protected function getDb() {
      if($this->db !== null) {
        return $this->db;
      }
      global $db;
      if(!isset($db)) {
        throw new EException(E_DEFAULT_DATABASE_NOT_DEFINED);
      }
      return $db;      
    }    
  }

  /**
  * Base class for logging
  */
  abstract class TCustomLog extends TObject {
    /** 
    * Date format 
    * for function date()
    *
    * @var string         
    */
    public $dateFormat = 'Y-m-d H:i:s';
    /** @var bool Is logging enabled */        
    public $enabled = true;
    /** 
    * Default entry type
    * E_NOTICE, E_WARNING, E_ERROR
    *
    * @var int
    */ 
    public $defaultType = E_NOTICE;
    /**
    * Stronger or equal will be logged
    * E_NOTICE, E_WARNING, E_ERROR
    *
    * @var int
    */
    public $filterLevel = E_NOTICE;

    /**
    * Write log entry text to output
    *
    * @param string Entry text
    * @return void 
    */
    protected abstract function logLine($entry_text);
    /**
    * Write new log entry
    *
    * @param string Entry text
    * @param int Entry type
    * @return void    
    */    
    public function log($entry_text,$type = null) {
      if(!$this->enabled) return;

      if(is_null($type)) {
        $type = $this->defaultType;
      } else {
        $type = errorLevel($type);       
      }
      if($type > $this->filterLevel) return;      
      switch($type) {
        case E_ERROR: $prefix = '[ERROR]'; break; 
        case E_WARNING: $prefix = '[WARNING]'; break;        
        default: $prefix = '[NOTICE]'; break;        
      }
      $this->logLine(date($this->dateFormat).' '.$prefix.': '.$entry_text);    
    }    
  }


  /**
  * File log
  */
  class TFileLog extends TCustomLog {
    /** @var string Filename of log file */
    protected $filename;
    /**
    * Constructor
    *
    * @param string Log file filename
    * @param int Rights for new file, if it doesn't exist yet (default 0666)
    * @return GLogFile
    */
    public function __construct($filename, $rights = 0666) {
      if(is_file($filename)) {
        if(!is_readable($filename) || !is_writable($filename)) {
          throw new EException('File must be both readable and writeable.');
        }
      } else {
        if(!@touch($filename)) {
          throw new EException("File couldn't be created.");          
        }
        chmod($filename,$rights);
      }
      $this->filename = $filename;
    }
    /**
    * Returns filename of log file
    *
    * @return string Filename
    */
    public function getFilename ( )  {
      return $this->filename;    
    }
    protected function logLine($entry_text) {
      $fd = fopen($this->filename, "a");
      fwrite($fd, $entry_text . "\n");
      fclose($fd);    
    }
    /**
    * Output log to browser
    *
    * @return void
    */    
    public function view() {
      readfile($this->filename);
    }
    /**
    * Clear log file
    *
    * @return void
    */
    public function clear() {
      fclose(fopen($this->filename,'w'));
    }
  }

  
  /**
  *
  */
  class TComponent extends TObject {
    private $validateResult = null;
    /** @var bool Allow sub-controls to be aligned in rows */
    public $allowRows = true;
    /** @var int DO NOT ACCESS DIRECTLY - Render dependant runtime flags */
    public $renderFlags = 0;    
    static protected $totalComponentCount = array();
    /** @var int ID of a row if this control is a part of TControlPager */
    public $pagerId;
    /** @var TComponent parent control  */
    public $parent = null;
    /** @var TPage parent page */
    public $page;
    /** @var TPanel parent panel */
    public $panel;
    /** @var TControlPager parent pager */
    public $pager;
    /** @var TRoad framework core */
    public $road;
    /** @var TComponent[] array of child components */        
    public $components = array();
    /** @var string[] array of request available methods' names */
    protected $actions = array();
    /** @var string Component name */
    public $name;
    /** @var TCustomRenderer explicitly defined renderer */
    protected $renderer = null;
    /** @var TCustomDatabase */    
    private $db = null;
    /** @var TForm Parent form if any */
    public $form = null;
    /** @var int Database load/save options */
    public $dbOptions = DB_BOTH;
    /** @var TDataSet */
    public $dataset = null;
    /** @event(TComponent $sender) User validation of the value (return True if valid) */
    public $onValidate = false;
    /** @event(TComponent $sender) Occurs when component has been initialized */
    public $onInitialized = false;    
    /** @event(TComponent $sender, $flags) Occurs when component is about to be rendered */
    public $onRender = false;
    /** @var int Automatically refresh control after selected interval */
    public $refreshInterval = null;
    /** @var bool Should the panel be instantly visible? */
    public $visible = true;
    /** @var bool Allow AJAX showing of this panel */
    public $dynamic = true;
    /** @var string JavaScript to be executed after the control is rendered */
    public $javaScript = '';
    /** @var string jQuery code to be executed after the control is rendered */
    public $jQuery = '';
    public $renderComponent = true;
    protected $errorMessage = false;
    protected $errorCode = false;
    
    public function setPagerId($value)  {
      $this->pagerId = $value;
      foreach($this->components as $c) {
        $c->setPagerId($value);
      }
    }
    public function __get($prop_name) {
      if(method_exists($this,'_get'.$prop_name)) {
        return $this->{'_get'.$prop_name}();
      } else if(property_exists($this,'_'.$prop_name)) {
        return $this->{'_'.$prop_name};
      } else if($component = $this->findComponent($prop_name,true)) {
        return $component;
      } else {
        throw new EException(E_UNDEFINED_PROPERTY,array($this->getClass().'::'.$prop_name));
      } 
    }    
    /**
    * Initialize component prior initializin others and parsing actions
    *
    * @return void
    */
    public function preInit() {
      foreach($this->components as $component) {
        $component->preInit();
      }
    }    
    /**
    * Add sub-component
    *
    * @return void
    */
    public function addComponent($component)  {
      $this->components[] = $component;
      $component->parent = $this;
      $component->panel = $this->panel;
      $component->page = $this->page;
      $component->road = $this->road;
      $component->form = $this->form;
      $component->pager = $this->pager;      
/*      $c = $this->parent;

      while($c !== null) {
        if($c->isInstanceOf('TPage')) {
          $this->page = $c;
        } elseif($c->isInstanceOf('TPanel')) {
          $this->panel = $c;
        } elseif($c->isInstanceOf('TRoad')) {
          $this->road = $c;
        } elseif($c->isInstanceOf('TForm')) {
          $this->form = $c;
        }
        $c = $c->parent;
        if(($c == null) || ($c == $c->parent)) {
          break;
        } 
      }*/
    }
    /**
    * Clone component
    *
    * @return TComponent
    */
    public function cloneMe($keepNames = false)  {
      /** @var TComponent */
      $res = clone $this;
      if(!$keepNames) {
        $res->setUniqueName();
      }
      $res->components = array();
      foreach($this->components as $c) {
        $x = $c->cloneMe($keepNames);
        $x->parent = $res;
        $res->components[] = $x;
      }
      return $res;
    }    
    /**
    * Remove the component
    *
    * @return void
    */
    public function delete() {
      if(($index = array_search($this,$this->parent->components)) !== false) {
        unset($this->parent->components[$index]);
      }

    }
    /**
    * Initialize component when all components have been created
    *
    * @return void
    */
    public function init() {
      foreach($this->components as $component) {
        $component->init();
      }
      $this->fireEvent('onInitialized');
    }    
    /**
    * Get unique identification string for this component
    *
    * @return string
    */
    public function getId() {
      return $this->name;
    }
    /**
    * Check if the values submited by user are valid
    *
    * @return bool
    */
    public function validate() {
      if($this->validateResult !== null) {
        return $this->validateResult;
      }
      $res = true;
      foreach($this->components as /** @var TComponent */ $component) {
        $res = $component->validate() && $res;
      }
      
      if($res) {
        $res = $this->fireEvent('onValidate');
        if(($res === null) || $res) {        
          $res = true; //return true;
        } else {
          $res = false; //return false;
        }
      } else {
        $res = false; //return false;
      }
      $this->validateResult = $res;
      return $res;

    }        
    /**
    * Returns true if an error has occured while saving to database
    *
    * @return bool
    */
    public function isErrorneous( ) {       
      if($this->errorMessage !== false) {
        return true;
      }
      foreach($this->components as /** @var TComponent */ $component) {
        if($component->isErrorneous()) {
          return true;
        }
      }
      return false;
    }
    /**
    * Returns the last error message
    *
    * @return string
    */
    public function getErrorMessage( )  {
      return $this->errorMessage;
    }
    /**
    * Returns the last error code
    *
    * @return int
    */
    public function getErrorCode( )  {
      return $this->errorCode;
    }
    /**
    * Set the last error message and code  
    *
    * @param string
    * @param int
    * @return void
    */
    public function setError($messageOrCode, $codeOrArgs = 0)  {
      if(is_int($messageOrCode)) {
        if(is_array($codeOrArgs)) {
          $args = $codeOrArgs;
        } else {
          $args = array();
        }
        $codeOrArgs = $messageOrCode;
        $messageOrCode = formatError($messageOrCode,$args,false);
      }
      $this->errorMessage = $messageOrCode;
      $this->errorCode = $codeOrArgs;    
    }
    /**
    * Remove error
    *
    * @return void
    */
    public function clearError()  {
      $this->errorMessage = false;    
    }
    /**
    * Returns current dataset
    *
    * @return TDataSet
    */
    public function getDataset()  {
      if($this->dataset === null) {
        throw new EException(E_NO_ACTIVE_DATASET,array($this->name)); 
      }
      return $this->dataset;
    }    
    /**
    * Occurs when dataset is assigned
    *
    * @return void
    */
    public function load() {}
    /**
    * Set dataset
    *
    * @return void
    */
    public function setDataset(TDataSet $dataset)  {
      $this->dataset = $dataset;
      foreach($this->components as $component) {
        $component->setDataSet($dataset);
      }
      $this->load();
    }             
    /**
    * Load component from user input
    *
    * @return void
    */
    public function loadInput(&$input) {
      foreach($this->components as /** @var TComponent */ $component) {
        $component->loadInput($input);
      }
      
    }
    /**
    * Load component from DataSet 
    *
    * @return void
    */
    public function loadFromDataset()  {
      foreach($this->components as /** @var TComponent */ $component) {
        if($component->dbOptions & DB_LOAD) {
          $component->loadFromDataSet();
        }
      }    
    }
    /**
    * Save component to DataSet 
    *
    * @return bool
    */
    public function saveToDataset()  {
      $res = true;
      foreach($this->components as /** @var TComponent */ $component) {
        if($component->dbOptions & DB_SAVE) {
          $res = $component->saveToDataSet() && $res;
        }
      }
      return $res;    
    }
    /**
    * Dataset has been saved to database.
    * If this component uses external table joined by to the main one,
    * it should save its data in this method. 
    *
    * @return bool
    */
    public function datasetSaved()  {
      $res = true;
      foreach($this->components as /** @var TComponent */ $component) {
        if($component->dbOptions & DB_SAVE) {
          $res = $res && $component->dataSetSaved();
        }
      }
      return $res; 
    }
    /**
    * Everything has been saved to database, update auto values 
    *
    * @return void
    */
    public function allSaved()  {
      foreach($this->components as /** @var TComponent */ $component) {
        if($component->dbOptions & DB_LOAD) {
          $component->allSaved();
        }
      }    
    }
    /**
    * Get database driver for this object
    *
    * @return TCustomDatabase    
    */
    protected function getDb() {
      if($this->db === null) {
        if(isset($this->parent)) {
          return $this->parent->getDb();
        } else {
          return parent::getDb();
        }
      } else {
        return $this->db;
      }
    }

    public function fireEvent($eventName, $arguments = array())  {
      if($this->isInstanceOf('TRoad')) {
        return parent::fireEvent($eventName,$arguments);
      }
      if($this->panel === null) {
        return parent::fireEvent($eventName,$arguments);
      }
      if(!isset($this->$eventName)) {
        throw new EException(E_UNDEFINED_EVENT,array($eventName));
      }
      $eventFunction = $this->$eventName;
      if($eventFunction) {
        if(substr($eventFunction,0,1) === '>') {
          return $this->evalEvent(substr($eventFunction,1),$arguments);
        }      
        if(!method_exists($this->panel,$eventFunction)) {
          throw new EException(E_EVENT_FUNCTION_NOT_FOUND,array($eventFunction,$eventName));
        }
        return call_user_func_array(array($this->panel,$eventFunction),array_merge(array($this),$arguments)); 
      } else {
        return null;
      }
    }
    /**
    * Set database driver for this object
    *
    * @param TCustomDatabase    
    */
    protected function setDb($db) {
      $this->db = $db;
    }
    /**
    * Get renderer suitable to render this component
    *
    * @return TCustomRenderer
    */    
    public function getRenderer() {
      if($this->renderer === null) {
        return $this->road->output->getRenderer($this);
      } else {
        return $this->renderer;
      }
    }
    /**
    * Explicitly set renderer for this component
    *
    * @param TCustomRenderer
    * @return void
    */    
    public function setRenderer(TCustomRenderer $renderer) {
      $this->renderer = $renderer;
    }    
    /**
    * Render component using default renderer 
    *
    * @return TComponent new instance
    */    
    public function render($flags = null) {
      $this->fireEvent('onRender',array('flags'=>$flags));
      $this->road->output->render($this,$flags);
      return false;
    }
    /**
    * Set component name to unique string
    *
    * @return void
    */
    public function setUniqueName() {
      if(isset(self::$totalComponentCount[$this->getClass()])) {
        self::$totalComponentCount[$this->getClass()] ++;
      } else {
        self::$totalComponentCount[$this->getClass()] = 1;
      }
      $this->name = strDelete($this->getClass(),0,1).self::$totalComponentCount[$this->getClass()];
    }
    /**
    * Constructor
    *
    * @var TComponent parent component
    * @return TComponent new instance
    */    
    public function __construct(TComponent $parent, $name = null, $properties = null) {    
      if(is_array($name)) {
        $properties = $name;
      }    
      if(is_string($name)) {
        $this->name = $name;
      } elseif(isset($propertis['name'])) {
        $this->name = $propertis['name'];
      } else {
        $this->setUniqueName();
      }      
      $this->actions['render'] = ACTION_AFTER_INPUT;
      if($parent != $this) {      
        $parent->addComponent($this);
      }
      if(CFG_USE_INDEXED_COMPONENT_LIST) {
        $this->road->componentIndex[$this->name] = $this;
      }
      if(is_array($properties)) {
        $this->assignArray($properties);
      }

    }
    public function findComponent($name, $deep = false) {
      if(CFG_USE_INDEXED_COMPONENT_LIST) {
        if(isset($this->road->componentIndex[$name])) {
          $res = $this->road->componentIndex[$name];
          $x = $res->parent;
          do {
            if($x == $this) {
              return $res;
            }
            $x = $x->parent;
          } while($deep && ($x !== null));
        }
      }
      foreach($this->components as $item) {
        if(sameText($item->name,$name)) {
          return $item;
        } elseif($deep) {
          $res = $item->findComponent($name,$deep);
          if($res) {
            return $res;
          }
        }
      }
      return false;
    }
  }

  /**
  *
  */
  class TControl extends TComponent {    
    /** @var mixed Width of the control */
    public $width = null;
    /** @var mixed Height of the control */
    public $height = null;
    /** @var string Additional custom CSS style */
    public $cssStyle = null;
    /** @var bool Should the control be visible? */
    public $visible = true;
    /** @var string Caption */
    public $caption = null;
    /** @var string Short description shown in the hint window */
    public $title = null;
    /** @var string Additional CSS class */
    public $cssClass = null;
    /** @var bool Is the control enabled? */
    public $enabled = true;
    /** @var bool Is the control read-only? */
    public $readOnly = false;
    /** @var string[] Array of additional and inherited CSS classes */
    public $additionalCssClasses = array();
    /** @var bool Break row and start in a new one */    
    public $newRow = true;
    /** @var bool Display caption */
    public $displayCaption = true;
    /** @var bool Allow AJAX enhancements of this control */
    public $dynamic = false;
    /** @var bool Set wheather the control should be focused */
    public $focused = false;
    
    public function validate() {
      $res = parent::validate();
      if(!$res) {
        $this->visible = true;
      }
      return $res;
    }
    /**
    * True if AJAX enhancements are enabled for this control
    * and AJAX is supported by the browser.
    *
    * @return bool
    */
    public function allowAjax( )  {
      return ($this->dynamic && $this->road->browser->supportsAjax()); 
    }
    /**
    * Return control content prepared JSON serialization
    *
    * @return mixed
    */
    public function toJson($flags = null) {}
    /**
    * Returns caption or name if no caption is set
    *
    * @return string
    */
    public function getCaption( )  {
      if($this->caption === null) {
        return $this->name;
      } else {
        return $this->caption;
      }
    }
  } 

  class TCollectionItem extends TObject {
    /** @var TObject Parent object */
    public $parent;
    /**
    *
    * @param TObject Parent object 
    */
    public function __construct($parent) {
      $this->parent = $parent;
    }
  }
  /**
  *
  */
  class TForm extends TControl {
    /** @var TSubmitButton */
    public $defaultButton = null;
    protected $reseted = false;
    /** @var int Form submit method */
    public $method = FORM_METHOD_POST;
    /** @var string Script to send form data to */    
    public $action = ':';
    /** @event(TComponent $sender) Form has been submited */
    public $onSubmited = false;
    /** @event(TComponent $sender) Form values have been restored to originals */
    public $onReseted = false;

    public function __construct(TComponent $parent, $name = null, $properties = null) {
      parent::__construct($parent,$name,$properties);
      $this->road->registerForm($this);
      $this->form = $this;
      $this->actions['reset'] = ACTION_BEFORE_INPUT;
    }
    public function loadInput(&$input) {
      if(!$this->reseted) {      
        parent::loadInput($input);
        $this->fireEvent('onSubmited');
      }
    }
    /**
    * Revert values to original
    *
    * @return void
    */
    public function reset() {
      $this->reseted = true;
      $this->fireEvent('onReseted');
    }    
  }

  /**
  *
  */
  abstract class TCustomRenderer extends TObject {
    /** @var TCustomOutput */
    public $output;
    /** @var TRoad */
    public $road;    
    /**
    * Render component's children
    *
    * @param TComponent
    * @return void
    */
    public function renderChildren(TComponent $component)  {
      foreach($component->components as $c) {
        if($c->renderComponent) {
          $c->render();
        }
      }
    }
    /**
    * Renderer component
    *
    * @param TComponent 
    * @return void
    */
    public function render(TComponent $component) {
      $this->renderHeader($component);
      $this->renderChildren($component);      
      $this->renderFooter($component);
    }
    /**
    * Render component header
    *
    * @param TComponent    
    * @return void
    */
    public function renderHeader(TComponent $component)  {}

    /**
    * Render component header
    *
    * @param TComponent    
    * @return void
    */    
    public function renderFooter(TComponent $component)  {}

    /**
    * @param TRoad ROAD core
    */
    public function __construct($road) {
      $this->road = $road;
    }
  }
 
  /**
  *
  */
  abstract class TCustomOutput extends TObject {
    /** @var int */
    public $type = OUTPUT_FULL;     
    protected $headersSent = false;
    /** @var string Buffered output */
    protected $buffer;
    /** @var bool Is output buffered? */
    public $buffered = false;    
    /** @var string Encoding (html charset) */
    #public $encoding = 'utf-8';
    protected $_encoding = 'iso-8859-1';
    protected $reencode = false;
    /** @var TRoad */
    public $road;
    /** @var TRenderer[] Renderers */
    protected $renderers;
    /** @var bool Deflate the output to save space */
    #public $deflate = false;    
    protected $_deflate = false;
    /** @var string Used deflate algorithm */
    #public $deflateType = 'gzip';
    protected $_deflateType = 'gzip';    
    /** @var string Content type */
    public $contentType = 'text/plain';
    /**
    * Redirect parent page
    *
    * @return bool Supported
    */
    public function ajaxRedirect($url)  {    
      return false;
    }
    /**
    * 
    *
    * @param Exception
    * @return bool
    */
    public function renderError($errno, $errstr, $errfile = '', $errline = 0, $errcontext = array()) {}
    /**
    * 
    *
    * @param Exception
    * @return bool
    */
    public function renderException($exception) {}
    /**
    * Reencode string to match output encoding if needed
    *
    * @param string
    * @return string
    */
    public function reencode($buffer) {
      if($this->reencode) {
        return iconv(CFG_ENCODING,$this->_encoding.'//TRANSLIT',$buffer);
      } else {
        return $buffer;
      }
    }
    /**
    * Returns buffer content when $buffered is true
    *
    * @return string
    */
    public function getBuffer( )  {
      return $this->buffer;
    }
    /**
    * A warning has been reported at this spot 
    *
    * @return void
    */
    public function warningReported($warning, $warningCount)  {
    }
    /**
    * Render triggered warnings and notices 
    *
    * @return void
    */
    public function renderWarnings($warnings)  {
      $this->write('<div id="roadWarnings">
');    
      foreach($warnings as $warning) {
        $res = '
<br />';
        if($warning[0] == E_NOTICE) {
          $res .= '<b>Notice:</b> ';
        } else {
          $res .= '<b>Warning:</b> ';
        }
        $res .= $warning[1].' in '.$warning[2].' on line '.$warning[3].'<br />
';
        $this->write($res);
                
      }
      $this->write('</div>');
    }
    
    /**
    * Is this output supported
    *
    * @return bool
    */
    public function isSupported() {
      return true;
    }
    /**
    * Send HTTP headers to the browser
    *
    * @return void
    */
    public function sendHeaders()  {
      if($this->headersSent) {
        return;
      }
      header('Content-Type: '.$this->contentType.'; charset='.$this->encoding);
      if($this->deflate) {
        header('Content-Encoding: '.$this->_deflateType);      
      }
      $this->headersSent = true;    
    }
    /**
    * Write data to output buffer
    *
    * @return void
    */
    protected function write($buffer) {
      if($this->buffered) {
        $this->buffer .= $buffer;
      } else {
        echo $this->reencode($buffer);
      }
    }
    protected function _setDeflate($value) {
      if($value) {        
        if(contentEncodingAccepted('gzip') && function_exists('gzencode')) {
          $this->_deflateType = 'gzip';
        } elseif(contentEncodingAccepted('deflate') && function_exists('gzcompress')) {
          $this->_deflateType = 'deflate';
        } else {
          $this->_deflate = false;        
          return;
        }
        $this->_deflate = true;
        $this->buffered = true;
      } else {
        $this->_deflate = false;
      }
    }
    protected function _setEncoding($value) {
      $value = strtolower($value);
      $this->_encoding = $value;      
      $this->reencode = !sameText(CFG_ENCODING,$value);
      if($this->reencode) {
        if(!encodingSupported($value)) {
          if(sameText($value,CFG_DEFAULT_ENCODING)) {
            throw new EException(E_ENCODING_NOT_SUPPORTED,array($value));
          } else {
            $this->_setEncoding(CFG_DEFAULT_ENCODING);
          }
        }
      }      
    }
    /**
    * @param TRoad ROAD core
    */
    public function __construct($road) {
      $this->road = $road;      
      if(charsetEncodingAccepted(CFG_ENCODING)) {
        $this->encoding = CFG_ENCODING;
      } else {
        global $_ACCEPT_CHARSET;
        $this->encoding = 'iso-8859-1';
        foreach($_ACCEPT_CHARSET as $charset => $q) {
          if(encodingSupported($charset)) {
            $this->encoding = $charset;
            break;
          }
        }
      } 
    }        
    /**
    * Get renderer for an object
    *
    * @param mixed Classname or object to render 
    * @return TRenderer
    */
    public function getRenderer($classNameOrObject) {

      if(is_object($classNameOrObject)) {
        $classNameOrObject = get_class($classNameOrObject);
      }

      $c = $classNameOrObject;
      while((strlen($c) > 0) && !sameStr($c,'TObject')) {
        if(isset($this->renderers[$c])) {
          $result = $this->renderers[$c];
          $result->output = $this;
          return $result;
        } else {
          $c = get_parent_class($c);
        }
      }
      triggerError(E_RENDERER_NOT_FOUND, array($classNameOrObject),E_USER_WARNING);
      return null;      
//      throw new EException(E_RENDERER_NOT_FOUND, array($classNameOrObject));            
    }

    /**
    * Register new renderer
    *
    * @param TRenderer
    * @param string Class name of the object to be rendered
    * @return void
    */
    public function registerRenderer(TCustomRenderer $renderer, $className)  {
      $this->renderers[$className] = $renderer;
      $renderer->output = $this;
    }

    /**
    * Render component using suitable renderer
    *
    * @param TComponent component to render
    * @return void
    */    
    public function render(TComponent $component,$flags = null) {
      $renderer = $this->getRenderer($component);
      if($renderer !== null) { 
        $this->getRenderer($component)->render($component,$flags);
      }
    }
    /**
    * Output the buffer, clear cache or whatever renderer needs to do
    *
    * @return void
    */
    public function flush() {
      if($this->buffered) {
        if($this->reencode) {
          $buffer = $this->reencode($this->buffer);
        } else {
          $buffer = &$this->buffer;
        }
        if($this->_deflate) {
          if($this->deflateType == 'deflate') {
            echo gzcompress($buffer);
          } else {      
            echo gzencode($buffer);
          }
        } else {
          echo $buffer;    
        }
      }
    }
  }

  class TPanel extends TComponent {
    /** @var bool */
    private $initialized = false;
    /** @var bool */
    public $modal = false;
    /** @event(TComponent $sender, $exclusive) Occurs when panel is shown through an action */
    public $onShow = false;    
    /** @event(TComponent $sender) Occurs when panel is hidden through an action */
    public $onHide = false;
    /**
    * Set visible to true
    *
    * @return void
    */
    public function show($exclusive)  {    

      if($exclusive) {
        foreach($this->parent->components as /** TPanel */ $c) {
          if(($c !== $this) && $c->isInstanceOf('TPanel')) {
            if($c->renderComponent) {
              $c->renderComponent = false;
              $c->fireEvent('onHide');
            }
          }
        }
      }
      if(!$this->renderComponent) {
        $this->renderComponent = true;
        $this->fireEvent('onShow',array('exclusive'=>$exclusive));
      }
      return true;    
    }
    public function __construct(TComponent $parent, $name = null, $properties = null) {
      parent::__construct($parent,$name,$properties);
      $this->panel = $this;
      $this->actions['show'] = ACTION_BEFORE_INIT;
    }
    public function init() {
//      if(!$this->visible) return;
      if($this->initialized) {
        return;
      } else {
        $this->initialized = true;
      }
      parent::init();
    }    
  }
 
  /**
  *
  */
  class TPage extends TComponent {
    /** @var string Page title */
    public $title = '';
    /** @var string Page description */
    public $description = '';
    /** @var string Page keywords */
    public $keywords = '';    
    /** @var bool */
    private $initialized = false;
    public function __construct(TComponent $parent, $name = null, $properties = null) {
      parent::__construct($parent,$name,$properties);
//      $this->road->page = $this;
      $this->page = $this;
    }
    public function init() {
      if($this->initialized) {
        return;
      } else {
        $this->initialized = true;
      }
      parent::init();
    }    
  }

  /**
  *
  */
  class TSessionSettings extends TObject {
    /** @var string Session name */
    public $name = 'roadSession';
    /** @var string URI where the session is valid */
    public $path = CFG_WEB_ROOT;
    /** @var int Max minutes of inactivity before session ends */
    public $lifetime = null;        
  } 
  /**
  * Login and security handling class
  */
  class TSecurity extends TObject {
    /** @var TRoad */
    public $road;
    private $inited = false;
    /** @var int Authentication type (AUTH_SESSION, AUTH_HTTP) */
    public $authenticationType = CFG_AUTHENTICATION_METHOD;
    /** @event(TSecurity $sender) Login check event */
    public $onCheckLogin = false;
    /** @event(TSecurity $sender, $username, $password) Select apropriate user according to the HTTP auth header */
    public $onHttpLogin = false;        
    /** @event(TSecurity $sender, $rightRequired) Occurs user tries to execute action he doen't have enough rights for */
    public $onDenied = false;
    /** @event(TSecurity $sender, $oldIp, $newIp) Occurs when client IP doesn't match first request's IP */
    public $onIpMismatch = false;
    /** @event(TSecurity $sender, $oldUA, $newUA) Occurs when client user-agent doesn't match first request's user-agent */
    public $onUserAgentMismatch = false;
    /** @var mixed Custom information about current user */
    public $userInfo = null;
    /** @var mixed[] Array of user rights */
    public $rights = array();
    /** @var int */
    public $options = SECURITY_CHECK_ALL;
    /** @var TSessionSettings Session settings */
    public $sessionSettings;    
    private $loggedIn = false;

    public function denyAccess() {
      $this->denied('restricted');
    }
    private function denied($operation)  {
      if($this->fireEvent('onDenied',array('rightRequired' => $operation)) === null) {
        echo '<div style="font-family: \'Arial\'; font-size: 12pt;">'.$this->road->output->reencode(LNG_DENIED).'</div>';
      }
      exit;
    }
    /**
    * Returns true if user can perform this operation
    *
    * @parama mixed integer or string value
    * @return bool
    */
    public function isAllowed($operation) {
      if(($operation == '/') || ($operation === 0)) {
        return $this->isLoggedIn();
      }

      foreach($this->rights as $right) {
        if(is_int($operation) && is_int($right)) {
          if($operation <= $right) {
            return true;
          }
        } elseif(is_string($operation) && is_string($right)) {

          if(sameText('/superadmin/',includeTrailingSlash($right)) || startsText(includeTrailingSlash($operation),includeTrailingSlash($right))) {
            return true;
          }        
        }
      }
      return false;
    }
    /**
    * Check if user has enough rights. If not, onDenied event is triggered 
    *
    * @return void
    */
    public function requireRight($operation)  {
      if(!$this->isAllowed($operation)) {
        $this->denied($operation);
      }    
    }
    /**
    * Login new user 
    *
    * @param mixed Custom user information
    * @param mixed Array or comma separated list of user rights         
    * @return void
    */
    public function login($userInfo, $rights = '/')  {
      if(is_string($rights)) {
        $rights = explode(',',$rights);
      } elseif(!is_array($rights)) {
        $rights = array($rights);
      }
 
      if($this->authenticationType = AUTH_SESSION) {
        $_SESSION['roadUserInfo'] = $userInfo;
        $_SESSION['roadUserRights'] = $rights;
        $this->userInfo = &$_SESSION['roadUserInfo'];
        $this->rights = &$_SESSION['roadUserRights'];
      } else {
        $this->userInfo = $userInfo;
        $this->rights = $rights;
      }
      $this->loggedIn = true;
    }
    /**
    * Logout new user 
    *
    * @return void
    */
    public function logout()  {
      $this->userInfo = null;
      if(isset($_SESSION['roadUserInfo'])) {
        unset($_SESSION['roadUserInfo']);
        unset($_SESSION['roadUserRights']);
      }
      $this->loggedIn = false;
    }
    /**
    * Is the user logged in 
    *
    * @return bool
    */    
    public function isLoggedIn() {
      return $this->loggedIn;
    }
    
    public function __construct($road) {
      $this->sessionSettings = new TSessionSettings();
      $this->road = $road;
    }
    
    /**
    * Load userinfo from session 
    *
    * @return void
    */    
    public function init() {
      if($this->inited) {
        return;
      }
      $this->inited = true;



      if(!$this->sessionSettings->lifetime !== null) {
        session_cache_expire($this->sessionSettings->lifetime);
      }
      $this->sessionSettings->lifetime = session_cache_expire();        
      session_set_cookie_params(0,$this->sessionSettings->path);
      $cookieParams = session_get_cookie_params(); 
      $this->sessionSettings->path = $cookieParams['path'];
  
      session_name($this->sessionSettings->name);
      $this->sessionSettings->name = session_name();
      if(CFG_ALLOW_URL_SESSION && isset($_GET[$this->sessionSettings->name])) {
        $_COOKIE[$this->sessionSettings->name] = $_GET[$this->sessionSettings->name];
      }
      if(!session_id()) {
        $er = roadErrorReporting(E_STRICT);
        session_start();
        roadErrorReporting($er);
      }
      
      if($this->authenticationType == AUTH_HTTP) {
        if(!$this->onHttpLogin) {
          throw new EException(E_NO_AUTH_EVENT_ASSIGNED);
        }
        if(isset($_SERVER['PHP_AUTH_USER']) && isset($_SERVER['PHP_AUTH_PW'])) {
          $this->fireEvent('onHttpLogin',array('username' => $_SERVER['PHP_AUTH_USER'],'password' => $_SERVER['PHP_AUTH_PW']));
        }
        if(!$this->isLoggedIn()) {
          header('WWW-Authenticate: Basic realm="'.$this->road->output->reencode(str_replace('"','',CFG_APPLICATION_NAME)).'"');
          header('HTTP/1.0 401 Unauthorized');
          header('Content-Type: text/html; charset='.$this->road->output->encoding);
          $this->denied('/');
        }
                

      } else {
        if(isset($_SESSION['roadUserInfo'])) {
          $this->userInfo = &$_SESSION['roadUserInfo'];
          $this->rights = &$_SESSION['roadUserRights'];          
          $this->loggedIn = true;
          $this->fireEvent('onCheckLogin');
        }
      }
    }
    
  }

  /**
  *
  */
  class TBrowser extends TObject {
    private $capsArray;
    /** @var string Operating system */
    public $platform;
    /** @var float Browser version */
    public $version;
    /** @var string Browser name */
    public $name;
    /** @var string HTTP_USER_AGENT string */
    public $userAgent;
    /** @var int Browser capabilities */
    public $caps = 0;
    /** @var string IP Address of the client */
    public $ip;
    /** @var int Timestamp of first request */
    public $firstRequest;
    /** @var bool Browser has been detected */
    public $browserDetected = false;
    /** @var float CSS version supported */
    public $cssVersion = 0;
    private function addCap($key,$value) {
      if(isset($this->capsArray[$key]) && (($this->capsArray[$key] == 'true') || ($this->capsArray[$key] == '1'))) {
        $this->caps = $this->caps | $value;
      }
    }
    public function supportsJavascript() {
      return ($this->caps & BROWSER_JAVASCRIPT) > 0;
    }
    public function supportsCss($version = 1) {
      return ($this->caps & BROWSER_CSS) && ($version <= $this->cssVersion);
    }    
    public function supportsAjax() {
      return ($this->caps & BROWSER_AJAX) > 0;
    }
    public function supportsActiveX() {
      return ($this->caps & BROWSER_ACTIVEX) > 0;
    }
    public function supportsCookies() {
      return ($this->caps & BROWSER_COOKIES) > 0;
    }
    public function supportsSession() {
      return CFG_ALLOW_URL_SESSION || $this->supportsCookies();
    }
    public function supportsFrames() {
      return ($this->caps & BROWSER_FRAMES) > 0;
    }
    public function supportsIFrames() {
      return ($this->caps & BROWSER_IFRAMES) > 0;
    }
    public function supportsTables() {
      return ($this->caps & BROWSER_TABLES) > 0;
    }
    public function supportsJava() {
      return ($this->caps & BROWSER_JAVA) > 0;
    }
    public function supportsVBScript() {
      return ($this->caps & BROWSER_VBSCRIPT) > 0;
    }                
    public function isMobileDevice() {
      return ($this->caps & BROWSER_MOBILE) > 0;
    }
    public function isCrawler() {
      return ($this->caps & BROWSER_CRAWLER) > 0;
    }
    public function isDesktop() {
      return ($this->caps & BROWSER_DESKTOP) > 0;
    }
    public function isSyndicateReader() {
      return ($this->caps & BROWSER_SYNDICATEREADER) > 0;
    }    
    public function __construct($caps) {
      $this->firstRequest = time();    
      $this->capsArray = $caps;
      $this->userAgent = $_SERVER['HTTP_USER_AGENT'];
      $this->name = $caps['browser'];
      $this->version = $caps['version'];
      $this->platform = $caps['platform'];      


      $this->addCap('ismobiledevice',BROWSER_MOBILE);
      $this->addCap('crawler',BROWSER_CRAWLER);
      $this->addCap('issyndicationreader',BROWSER_SYNDICATEREADER);
      if(!($this->caps & (BROWSER_MOBILE | BROWSER_CRAWLER | BROWSER_SYNDICATEREADER))) {
        $this->caps = $this->caps | BROWSER_DESKTOP;
      }
      $this->addCap('tables',BROWSER_TABLES);
      $this->addCap('frames',BROWSER_FRAMES);
      $this->addCap('iframes',BROWSER_IFRAMES);


      $this->addCap('supportscss',BROWSER_CSS);
      if(($this->caps & BROWSER_CSS) && isset($caps['cssversion'])) {
        $this->cssVersion = $caps['cssversion'];
        if($this->cssVersion >= 2) {
          $this->caps = $this->caps | BROWSER_CSS2;
          if($this->cssVersion >= 3) {
            $this->caps = $this->caps | BROWSER_CSS3;        
          }
        }
      }
      if(!CFG_REAL_BROWSER_DETECT) {
        $this->addCap('cookies',BROWSER_COOKIES);
        $this->addCap('javascript',BROWSER_JAVASCRIPT);
        if($this->supportsJavascript()) {
          $this->caps = $this->caps | BROWSER_AJAX;
        }
      } else {
        if(isset($_COOKIE['roadTestCookie'])) {
          $this->addCap('cookies',BROWSER_COOKIES);
        }
      }
      $this->addCap('vbscript',BROWSER_VBSCRIPT);
      $this->addCap('javaapplets',BROWSER_JAVA);
      $this->addCap('activexcontrols',BROWSER_ACTIVEX);


    }
    /**
    * Output human readable information about browser
    *
    * @return void
    */
    public function info($return = false)  {
      $code = '<big><strong>'.htmlspecialchars($this->name).'</strong> version '.htmlspecialchars($this->version)."</big><br/>\n";
      $code .= "<small>".htmlspecialchars($this->userAgent)."</small><br/>\n";
      $dt = new TDateTime(null);
      $dt->setRaw($this->firstRequest);
      
      $code .=
        ($this->isCrawler() ? 
        '<strong>crawler</strong>' : 
          ($this->isCrawler() ? '<strong>syndicate reader</strong>' :        
        'browser'))." running on ".($this->isMobileDevice() ? '<strong>mobile</strong>' : '')." <i>".htmlspecialchars($this->platform)."</i> connected from <i>".htmlspecialchars($this->ip)."</i> at <i>".htmlspecialchars($dt->getAsRfc2822())."</i><br/>\n";
      $code .= "<ul>\n";
      $code .= "<li>Javascript: ".($this->supportsJavascript() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Css: ".($this->supportsCss() ? '<strong>'.$this->cssVersion.'</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Ajax: ".($this->supportsAjax() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>ActiveX: ".($this->supportsActiveX() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Cookies: ".($this->supportsCookies() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Session: ".($this->supportsSession() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Frames: ".($this->supportsFrames() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>IFrames: ".($this->supportsIFrames() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Tables: ".($this->supportsTables() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>Java: ".($this->supportsJava() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "<li>VBScript: ".($this->supportsVBScript() ? '<strong>OK</strong>' : '<small>N/A</small>')."</li>\n";
      $code .= "</ul>\n";
      if($return) {
        return $code;
      } else {
        echo $code;
      }                  
    }    
  } 
  /**
  *
  */
  class TRoad extends TComponent {
    public $componentIndex = array();
    /** @var bool */
    protected $initialized = false;
    protected $actionCache = array(ACTION_BEFORE_INPUT=>array(),ACTION_AFTER_INPUT=>array());
    /** @var bool */
    protected $actionsParsed = false;
    /** @var bool */
    protected $formsHandled = false;    
    /** @var array[] Triggered warnings and notices */
    protected $warnings = array();
    /** @var TSecurity */
    public $security;
    /** @var float Time when the execution started */
    public $executionStartTime;
    /** @var TCustomOutput */
    public $output;
    /** @var TBrowser Browser information */
    public $browser;
    /** @var TForm[] */
    protected $forms = array();
    private $ajaxed = false;
    /** @var string[] */
    public $libraries = array();
    /** @var string[] */
    protected $messages = array();
    private $storesMessagesInSession = false;
    
    /**
    *
    *
    * @return void
    */
    public function useLibrary($library)  {
      global $CFG_LIBRARIES;
      if(isset($CFG_LIBRARIES[$library])) {
        if(!isset($this->libraries[$library])) {
          $this->libraries[$library] = $CFG_LIBRARIES[$library];
          foreach($this->libraries[$library] as $file) {
            $path_info = pathinfo($file);
            if(isset($path_info['extension']) && sameText($path_info['extension'],'php')) {
              require_once(CFG_ROAD_DIR.'/'.$file);
            }
          }
        }
      } else {
        throw new EException(E_LIBRARY_NOT_DEFINED,array($library));
      }
    }
    public function addComponent($component) {
      if($component->isInstanceOf('TPage')) {
        $this->page = $component;
      }
      parent::addComponent($component);
    }
    /**
    * Returns true, if mobile version of output is enabled
    *
    * @return bool
    */
    public function isMobileVersion()  {
      return ($this->output->type & OUTPUT_MOBILE) > 0;
    }
    /**
    * Has the request been made through ajax
    *
    * @return bool
    */
    public function isAjaxed()  {
      return $this->ajaxed;
    }
    /**
    * True if there is any message [of such types] queued
    *
    * @param int One or more types 
    * @return bool
    */
    public function hasMessages($ofTypes = null)  {
      if($ofTypes === null) {
        return (count($this->messages) > 0);
      } else {
        foreach(array_keys($this->messages) as $type) {
          if($type & $ofTypes) {
            return true;
          }
        }
        return false;
      }
    }
    /**
    * Add a new message
    *
    * @param string Message text
    * @param int Message type    
    * @return void
    */
    public function queueMessage($text, $type) {
      if(isset($this->messages[$type])) {
        $this->messages[$type] = array();
      }
      $this->messages[$type][] = $text;
      if($this->storesMessagesInSession) {
        $_SESSION['rMessages'] = $this->messages;
      }      
    }
    /**
    * Get all messages of selected type
    *
    * @param int Type
    * @return string[]
    */
    public function dequeueMessages($type)  {
      if(isset($this->messages[$type])) {
        return $this->messages[$type];
        unset($this->messages[$type]);
        if($this->storesMessagesInSession) {
          unset($_SESSION['rMessages'][$type]);        
//          $_SESSION['rMessages'] = $this->messages;
        }
      } else {
        return array();
      }
    }
    /**
    * Clear all messages
    *
    * @return void
    */
    public function clearMessages() {
      $this->messages = array();
      if($this->storesMessagesInSession) {
        $_SESSION['rMessages'] = array();
      }          
    }
    /**
    * Register new TForm component
    *
    * @param TForm
    * @return void
    */
    public function registerForm(TForm $form)  {
      $this->forms[] =  $form;
    }
    
    /**
    * Handle form data submited by the user
    *
    * @return bool Have user posted any data? 
    */
    protected function handleForms() {
      if($this->formsHandled) {
        return false;
      } else {
        $this->formsHandled = true;
      }
      $forms = array();
      if(isset($_POST['roadFormSubmited']) && is_array($_POST['roadFormSubmited'])) {
        foreach($_POST['roadFormSubmited'] as $form) {
          $forms[$form] = &$_POST;
        }
      }
      if(isset($_GET['roadFormSubmited']) && is_array($_GET['roadFormSubmited'])) {
        foreach($_GET['roadFormSubmited'] as $form) {
          unset($forms[$form]);
          $forms[$form] = &$_GET;
        }
      }
      if(count($forms) < 1) {
        return false;
      }
      
      foreach(array_keys($forms) as $formName) {
        $input = &$forms[$formName];
        foreach($this->forms as /** @var TForm */ $form) {
          if(sameText($form->name,$formName)) {
            $form->loadInput($input);
            continue 2;
          }
        }
      }
    }
    /**
    * Used in error handling function
    *
    * @return bool
    */
    public function reportWarning($errorType, $message, $file, $line)  {
      $w = array($errorType,$message,$file,$line);
      $this->warnings[] = $w;
      $cnt = count($this->warnings);
      if(CFG_ADVANCED_ERROR_HANDLER && CFG_SHOW_WARNING_OCCURENCES && !empty($this->output)) { 
        return $this->output->warningReported($w,$cnt);
      }
      return null;
    }
    /**
    * Render warnings (if any) using default output
    *
    * @return void
    */
    public function renderWarnings()  {
      if(CFG_ADVANCED_ERROR_HANDLER && (count($this->warnings) > 0) && !empty($this->output)) {
        $this->output->renderWarnings($this->warnings);
      }
    }
    public function redirect($uri, $permanent = false) {
      $uri = absoluteUri($uri,true);
      if(!$this->storesMessagesInSession) {
        $query = array();
        foreach($this->messages as $type => $messages) {
          if(!isset($query['rMessages'])) {
            $query['rMessages'] = array();
          }
          foreach(array_keys($messages) as $key) {
            $messages[$key] = urlencode($messages[$key]);
          }
          $query['rMessages'][$type] = implode(',',$messages);
        }
        $uri = changeUrlQuery($uri,$query);
      }
      if(headers_sent()) {
        triggerError(E_CANNOT_REDIRECT,array($uri),E_USER_WARNING);
      } else {
        if($this->isAjaxed() && ($this->output !== null)) {        
          $this->output->ajaxRedirect($uri);
          exit;
        }
        header('Location: '.$uri,true,($permanent ? 301 : 302));
      }
      echo '<div style="font-family: \'Arial\'; font-size: 12pt"><p>'.htmlspecialchars(LNG_MANUAL_REDIRECTION).'</p>'
        .'<a href="'.htmlspecialchars($uri).'">'.htmlspecialchars($uri).'</a></div>';
      exit;
    }
    private function parseAction($actionString, &$path, &$method, &$arguments) {
      if(preg_match('/^([a-zA-Z0-9_\-]+)\((.*)\)$/',$actionString,$regs) == 0) return false;
      $path = explode('-',$regs[1]);
      $method = array_pop($path);             

      $s = $regs[2]; 
      $l = strLength($s);
      $i = 0;

      $nextExpected = false;
      $string = false;
      $quote = false;      
      $escape = false;
      $argument = 0;
      $arguments = array();
      $arguments[0] = '';
      $variables = array();
      $number = false;
      $space = false;
      $point = false;
      while($i < $l) {
        $x = $s[$i];
        if($string || $quote) {
          if($escape) {
            switch($x) {
              case 'n':
                $x = "\n";              
                break;
              case 'r':
                $x = "\r";
                break;
              case 't':
                $x = "\t";
                break;                
              default:
                break;  
            }
            $arguments[$argument] .= $x;
            $escape = false;
          } else {
            switch($x) {
              case "\\":
                $escape = true;              
                break;
              case "'":
                if($string) {
                  $string = false;
                  break;
                }
              case "`":
                if($quote) {              
                  $quote = false;
                  break;
                }                
              default:
                $arguments[$argument] .= $x;
                break;
            }
          }          
        } else {
          switch($x) {

            case "'":
              $string = true;              
              break;
            case ".":
              if($number) {
                $arguments[$argument] .= $x;
                $point = true;                              
              } else {
                return false;
              }              
              break;              
            case "`":
              $quote = true;
              $variables[$argument] = true;              
              break;              
            case ',':
              if($point) {
                return false;
              }
              $argument ++;
              $string = false;
              $quote = false;
              $space = false;                            
              $number = false;              
              break;
            case ' ':
              if($number) {
                $space = true;
              }
              break;              
            default:
              if(is_numeric($x) && (strLength($arguments[$argument]) == 0) || $number) {
                if($space) {
                  return false;
                }
                $number = true;
                $point = false;
                $arguments[$argument] .= $x;
              } else {             
                return false;
              }  
          }
        }
        $i++;
      }
      if($point) {
        return false;
      }
      foreach($variables as $key => $item) {
        if(isset($_POST[$arguments[$key]])) {
          $arguments[$key] = $_POST[$arguments[$key]]; 
        } else {
          throw new EException(E_ACTION_VARIABLE_NOT_FOUND,array($arguments[$key]));
        }              
      }
      return true;                            
    }
    /**
    * Returns the duration for which the script runs
    *
    * @return void
    */
    public function getExecutionTime()  {
      return microtime(true)-$this->executionStartTime;    
    }  
    public function __construct() {
      parent::__construct($this);
      $this->ajaxed = (isset($_SERVER['HTTP_X_REQUESTED_WITH']) && ($_SERVER['HTTP_X_REQUESTED_WITH'] == 'XMLHttpRequest'));      
      $this->parent = null;
      $this->page = null;
      $this->road = $this;
      $this->security = new TSecurity($this);
    }
    private function executeActions($type) {
 
      if(isset($this->actionCache[$type])) {
        $x = $this->actionCache[$type];      
        unset($this->actionCache[$type]);  
        foreach($x as $action) {
          if(call_user_func_array(array($action[0],$action[1]),$action[2]) !== true) {
            $this->output->flush();
            exit;
          }
        }
      }
    }
    /**
    * Adds action to actionCache for execution
    *
    * @return void
    */
    public function addAction($type, TComponent $component, $method, $arguments = array()) {
      $this->actionCache[$type][] = array($component,$method,$arguments);
    }
    private function parseActions()  {
      if($this->actionsParsed) {
        return false;
      } else {
        $this->actionsParsed = true;
      }
      $actions = array();
      foreach(array_keys($_GET) as $action) {
        if(startsStr(':',$action)) {
          $actions[] = strDelete($action,0,1);
        }
      }
      foreach(array_keys($_POST) as $action) {
        if(startsStr(':',$action)) {
          $actions[] = strDelete($action,0,1);
        }
      }
      if(isset($_GET['rAction'])) {
        if(is_array($_GET['rAction'])) {
          $actions = array_merge($actions,$_GET['rAction']);
        } else {
          $actions[] = $_GET['rAction'];
        }
      }      
      if(isset($_POST['rAction'])) {
        if(is_array($_POST['rAction'])) {
          $actions = array_merge($actions,$_POST['rAction']);
        } else {
          $actions[] = $_POST['rAction'];
        }
      }
      
      if(count($actions) > 0) {

        foreach($actions as $action) {
          if(!$this->parseAction($action,$path,$method,$arguments)) {
            throw new EException(E_ACTION_SYNTAX_ERROR,array($action));
          }
          /** @var TComponent */
          $component = $this->page;
          
          foreach($path as $item) {
            if(!($component = $component->findComponent($item,true))) {
              throw new EException(E_COMPONENT_NOT_FOUND,array($item));
            }        
          }
          if(isset($component->actions[$method]) && method_exists($component,$method)) {
            $this->addAction($component->actions[$method],$component,$method,$arguments);
          } else {
            throw new EException(E_METHOD_NOT_FOUND,array($method));
          }
        }
        return true;
      } else {
        return false;
      }
          
    }
    /**
    * Initialize components
    *
    * @return void
    */
    public function init( )  {
      if($this->initialized) {
        return true;
      } else {
        $this->initialized = true;
      }
      $this->security->init();
      if(!session_id()) {
        session_start();          
      }
      if(!empty($_SERVER['HTTP_CLIENT_IP'])) {
        $ip = $_SERVER['HTTP_CLIENT_IP'];
      } elseif(!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
        $ip = $_SERVER['HTTP_X_FORWARDED_FOR'];
      } else {
        $ip = $_SERVER['REMOTE_ADDR'];
      }      
      if(isset($_SESSION['rBrowser'])) {
        $this->browser = $_SESSION['rBrowser'];
        if(($this->security->options & SECURITY_CHECK_IP) && ($this->browser->ip <> $ip)) {
          if($this->security->fireEvent('onIpMismatch',array('oldIp' => $this->browser->ip, 'newIp' => $ip)) === null) {
            session_destroy();
            if($this->isAjaxed()) {
              exit;
            } else {
              $this->redirect(':');
            }
          }
        }
        if(($this->security->options & SECURITY_CHECK_USERAGENT) && ($this->browser->userAgent <> $_SERVER['HTTP_USER_AGENT'])) {
          if($this->security->fireEvent('onUserAgentMismatch',array('oldUA' => $this->browser->userAgent, 'newUA' => $_SERVER['HTTP_USER_AGENT'])) === null) {
            session_destroy();
            if($this->isAjaxed()) {
              exit;
            } else {
              $this->redirect(':');
            }
          }
        }        
      }
      if(!isset($_SESSION['rBrowser'])) {
        $this->browser = new TBrowser(getBrowserCapabilities());
        $_SESSION['rBrowser'] = &$this->browser;        
      }
      $this->browser->ip = $ip;  
      if(CFG_REAL_BROWSER_DETECT && isset($_GET['rBrowserCaps'])) {
        $caps = $_GET['rBrowserCaps'];
        if(!is_string($caps)) {
          $caps = '000';
        }
        $this->browser->browserDetected = true;

        if(substr($caps,0,1)) {
          $this->browser->caps = $this->browser->caps | BROWSER_JAVASCRIPT;
        }
        if(substr($caps,1,1)) {
          $this->browser->caps = $this->browser->caps | BROWSER_AJAX;
        }
        if(substr($caps,2,1) || isset($_COOKIE['roadTestCookie'])) {
          $this->browser->caps = $this->browser->caps | BROWSER_COOKIES;
        }
        if(CFG_REDIRECT_AFTER_BROWSER_DETECTION && $this->browser->supportsSession()) {
          $query = array('rBrowserCaps'=>false);
          if(CFG_ALLOW_URL_SESSION && !$this->browser->supportsCookies() && ($id = session_id())) {
            $query[$this->security->sessionSettings->name] = $id;
          } else {
            $query[$this->security->sessionSettings->name] = false;          
          }
          $this->redirect(changeUrlQuery(absoluteUri(':'),$query));
        }
      }
      
      $output = 'full';
      if($this->browser->isMobileDevice()) {
        $output = 'mobile';      
      }
      if(isset($_GET['rDefaultOutput']) && $this->browser->supportsSession()) {
        $_SESSION['rDefaultOutput'] = $_GET['rDefaultOutput'];
      }
      if(isset($_POST['rDefaultOutput']) && $this->browser->supportsSession()) {      
        $_SESSION['rDefaultOutput'] = $_POST['rDefaultOutput'];
      }
      if(isset($_SESSION['rDefaultOutput'])) {
        $output = $_SESSION['rDefaultOutput'];
      }

      if(isset($_GET['rOutput'])) {
        $output = $_GET['rOutput'];
      }  
      if(isset($_POST['rOutput'])) {
        $output = $_POST['rOutput'];
      }
      
      if((preg_match('/[a-z0-9]/i',$output) === 0) || !is_file(CFG_ROAD_DIR.'/output/'.$output.'/render-set.php')) {
        throw new EException(E_NO_OUTPUT_HANDLER,array($output));
      }
      require_once(CFG_ROAD_DIR.'/output/'.$output.'/render-set.php');

      if(is_file(CFG_ROAD_DIR.'/template/'.$output.'/template.php')) {
        require_once(CFG_ROAD_DIR.'/template/'.$output.'/template.php');
      }

      setcookie('roadTestCookie', '1', 0, includeTrailingSlash(CFG_WEB_ROOT));      

      if($this->browser->supportsSession()) {
        if(!session_id()) {
          session_start();
        }
        if(isset($_SESSION['rMessages'])) {
          $this->messages = $_SESSION['rMessages'];
        }
        $this->storesMessagesInSession = true;
      }
      if(isset($_GET['rMessages']) && is_array($_GET['rMessages'])) {
      
        foreach($_GET['rMessages'] as $type => $messages) {
          if(!isset($this->messages[$type])) {
            $this->messages[$type] = array();
          }
          $messages = explode(',',$messages);
          foreach($messages as $msg) {
            $this->messages[$type][] = urldecode($msg);
          }
        }
      }
      $this->output->sendHeaders();
      if(CFG_REAL_BROWSER_DETECT
        && !CFG_FULL_PAGE_FOR_UNDETECTED_BROWSER
        && !$this->browser->browserDetected
        && $this->output->isInstanceOf('TXhtmlOutput')
      ) {
        $this->page->render('undetectedBrowser');
        exit;
      }
            
    }
    
    private function internalStart() {
      $this->init();
      parent::preInit();
      $this->parseActions();
      $this->executeActions(ACTION_BEFORE_INIT);
      parent::init();      
      $this->executeActions(ACTION_BEFORE_INPUT);
      $this->handleForms();
      $this->executeActions(ACTION_AFTER_INPUT);
    }
    
    /**
    * Execute framework core
    *    
    * @return void
    */
    public function run()  {
      $this->internalStart();
      $this->page->render();
      $this->renderWarnings();
      $this->output->flush();
      exit;      
    }
    /**
    * Send headers and render file header
    *
    * @return void
    */
    public function start()  {
      $this->internalStart();
      $this->output->sendHeaders();
      $this->page->getRenderer()->renderHeader($this->page);
    }
    /**
    * Render file footer and stop script processing
    *
    * @return void
    */
    public function stop()  {
      $this->renderWarnings();
      $this->page->getRenderer()->renderFooter($this->page);
      $this->output->flush();
      exit();
    }
  } 

  /**
  * Base class for format objects 
  */
  abstract class TCustomValue extends TObject {
    /** @var TObject Parent object to be notified by the valueChanged(TCustomValue $value) when the value changes */
    public $parent;
    /** @var mixed Raw value stored for formating */
    private $values = array(0 => null);
    /** @var string Last error message */
    protected $errorMessage = '';
    protected $errorCode = 0;    
    /** @var bool If true convertor will issue warnings if needed*/
    public $verbose = true;
    /** @var bool If true convertor will trow exceptions instead of warnings */
    public $strict = false;    
    /** @var string Example of valid value */
    protected $inputExample = '';
    private $tempMode = 0;
    /**
    * Start temp mode
    *
    * @return int
    */
    public function startTempMode() {
      $this->values[$this->tempMode+1] = $this->values[$this->tempMode];
      $this->tempMode ++;
      return $this->tempMode;
    }
    /**
    * End temp mode
    *
    * @return int
    */
    public function endTempMode() {
      $this->tempMode --;
      if($this->tempMode < 0) {
        $this->tempMode = 0;
      }
      return $this->tempMode;
    }
    /**
    * Is the value NULL?
    *
    * @return bool
    */
    public function isNull() {
      return ($this->values[$this->tempMode] === null);
    }
    /**
    * Get example of valid value
    *
    * @return string
    */
    public function getInputExample()  {
      return $this->inputExample;    
    }
    /**
    *
    *
    * @return bool
    */
    protected function setError($messageOrCode,$arguments = array())  {    
      if(is_int($messageOrCode)) {
        $this->errorMessage = formatError($messageOrCode,$arguments);
        $this->errorCode = $messageOrCode;
      } else {
        $this->errorMessage = $messageOrCode;
        $this->errorCode = 0;
      }      
      if($this->strict) {
        throw new EException($this->errorMessage,$this->errorCode);
      } elseif($this->verbose) {
        triggerError(formatError($messageOrCode,$arguments,true),E_USER_WARNING);
      }
      return false;    
    }
    /**
    * Get value formated for SQL queries
    *
    * @return string
    */
    public function getAsSql() {
      return $this->getRaw();
    }

    /**
    * Get value formated and escaped for SQL queries
    *
    * @return string
    */
    public function getAsSqlEscaped() {
      return $this->getDb()->escape($this->getAsSql());
    }
    /**
    * Get value formated, escaped and quoted for SQL queries
    *
    * @return string
    */
    public function getAsSqlQuoted() {
      if($this->isNull()) {
        return 'NULL';
      } else {
        return $this->getDb()->quote($this->getAsSql());
      }
    }
        
    /**
    * Set value formated as recieved from SQL query result
    *
    * @param string 
    * @return bool Is the value valid
    */
    public function setAsSql($value) {
      return $this->setRaw($value);
    }

    /**
    * Get value formated for input edits
    *
    * @return string
    */
    public function getAsInput() {
      return $this->getRaw();
    }
    
    /**
    * Set value formated as recieved from input edits
    *
    * @param string 
    * @return bool Is the value valid
    */
    public function setAsInput($value) {
      return $this->setRaw($value);    
    }

    /**
    * Get value formated for direct output
    *
    * @return string
    */
    public function getAsDisplay() {
      return $this->getRaw();
    }
    
    /**
    * Set value formated as beeing displayed
    *
    * @param string 
    * @return bool Is the value valid
    */
    public function setAsDisplay($value) {
      return $this->setRaw($value);    
    }
        
    /**
    * Set value in native PHP format
    *
    * @param mixed    
    * @return bool Is the value valid
    */
    public function setRaw ($value)  {
      if($this->values[$this->tempMode] === $value) {
        return true;
      }
      $old = $this->values[$this->tempMode];
      $this->values[$this->tempMode] = $value;
            
      if(($this->tempMode == 0)
        &&
        ($this->parent !== null)
        &&
        method_exists($this->parent,'valueChanged')
      ) {
        $this->parent->valueChanged($this);
      }           
      return true;    
    }

    /**
    * Get value in native PHP format
    *
    * @return mixed
    */
    public function getRaw() {
      return $this->values[$this->tempMode];
    }

    /**
    * Get last error message
    *
    * @return string
    */    
    public function getLastError() {
      return $this->errorMessage;
    }

    /**
    * Get last error message
    *
    * @return string
    */    
    public function getLastErrorNo() {
      return $this->errorCode;
    }

    /**
    * Clear the value
    *
    * @return void
    */    
    public function clear() {
      $this->setRaw(null);
    }
    /**
    *
    * @param TObject Parent object to be notified via valueChanged(TCustomValue $value) method
    */
    public function __construct($parent = null) {
      $this->parent = $parent;
      $this->clear();
    }
  }
  
  /**
  *
  */
  class TConvertor extends TObject {
    /** @var TCustomValue[] */
    protected $instances = array();
    protected function getInstance($format) {
      if(!isset($this->instances[$format])) {
        $s = 'T'.$format;
        if(!(class_exists($s) && is_subclass_of($s,'TCustomValue'))) {
          throw new EException(E_FORMAT_CLASS_DOES_NOT_EXIST,array($format,strtoupper($s)));
        }
        $this->instances[$format] = new $s();  
      }
      return $this->instances[$format];
    }
    /**
    * Convert value from SQL to display format
    *
    * @param string
    * @param string    
    * @return string
    */    
    public function sqlToDisplay($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsSql($value)) {
        $instance->clear();
      }
      return $instance->getAsDisplay();
    }
    /**
    * Convert value from SQL to input format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function sqlToInput($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsSql($value)) {
        $instance->clear();
      }
      return $instance->getAsInput();
    }
    /**
    * Convert value from SQL to native PHP format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function sqlToRaw($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsSql($value)) {
        $instance->clear();
      }
      return $instance->getRaw();
    }
    /**
    * Convert value from input to SQL format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function inputToSql($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsInput($value)) {
        $instance->clear();
      }
      return $instance->getAsSql();
    }
    /**
    * Convert value from input to SQL escaped format 
    *
    * @param string
    * @param string
    * @return string
    */    
    public function inputToSqlEscaped($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsInput($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlEscaped();
    }
    /**
    * Convert value from input to SQL quoted format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function inputToSqlQuoted($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsInput($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlQuoted();
    }
    /**
    * Convert value from input to display format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function inputToDisplay($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsInput($value)) {
        $instance->clear();
      }
      return $instance->getAsDisplay();
    }
    /**
    * Convert value from input to native PHP format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function inputToRaw($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsInput($value)) {
        $instance->clear();
      }
      return $instance->getRaw();
    }
    /**
    * Convert value from display to SQL format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function displayToSql($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsDisplay($value)) {
        $instance->clear();
      }
      return $instance->getAsSql();
    }
    /**
    * Convert value from display to SQL escaped format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function displayToSqlEscaped($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsDisplay($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlEscaped();
    }
    /**
    * Convert value from display to SQL quoted format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function displayToSqlQuoted($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsDisplay($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlQuoted();
    }
    /**
    * Convert value from display to input format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function displayToInput($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsDisplay($value)) {
        $instance->clear();
      }
      return $instance->getAsInput();
    }
    /**
    * Convert value from display to native PHP format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function displayToRaw($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setAsDisplay($value)) {
        $instance->clear();
      }
      return $instance->getRaw();
    }
    /**
    * Convert value from native PHP format to SQL format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function rawToSql($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setRaw($value)) {
        $instance->clear();
      }
      return $instance->getAsSql();
    }
    /**
    * Convert value from native PHP format to SQL escaped format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function rawToSqlEscaped($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setRaw($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlEscaped();
    }
    /**
    * Convert value from native PHP format to SQL quoted format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function rawToSqlQuoted($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setRaw($value)) {
        $instance->clear();
      }
      return $instance->getAsSqlQuoted();
    }    
    /**
    * Convert value from native PHP format to input format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function rawToInput($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setRaw($value)) {
        $instance->clear();
      }
      return $instance->getAsInput();
    }
    /**
    * Convert value from native PHP format to display format
    *
    * @param string
    * @param string
    * @return string
    */    
    public function rawToDisplay($format, $value) {
      /** @var TCustomValue */
      $instance = $this->getInstance($format);
      if(!$instance->setRaw($value)) {
        $instance->clear();
      }
      return $instance->getAsDisplay();
    }
  }

  class THiddenData extends TComponent {
    /** @var mixed Data to store */
    public $data;
    public function loadInput(&$input) {
      if(isset($input[$this->getId()])) {
        $this->data = $input[$this->getId()];
      }
    }    
  }

  $convertor = new TConvertor(); 
?>