<?php

  class TDateTime extends TCustomValue {
    /** @var string strftime format used for direct output */
    public $displayFormat = LNG_DATE_TIME_FORMAT;
    /** @var string strftime format used for input edits */
    public $inputFormat = LNG_DATE_TIME_FORMAT;    
    /** @var string strftime format used for sql queries */
    public $sqlFormat = CFG_SQL_DATE_TIME_FORMAT;
    /** @var int Minimal time stamp allowed */
    public $minValue = null;
    /** @var int Maximal time stamp allowed */
    public $maxValue = null;
    /** @var int Allowed date time values */
    public $allowed = DATETIME_BOTH;
    /**
    * Returns $allowed or $minValue or false if no boundary set
    *
    * @return mixed
    */
    public function getMinValue()  {
      if(($this->minValue === null) && !($this->allowed & DATETIME_PAST)) {
        return time();
      } else {
        return false;
      }
    }
    /**
    * Returns $allowed or $maxValue or false if no boundary set
    *
    * @return mixed
    */
    public function getMaxValue()  {
      if(($this->maxValue === null) && !($this->allowed & DATETIME_FUTURE)) {
        return time();
      } else {
        return false;
      }
    }
    public function getInputExample()  {
      return strftime($this->inputFormat);    
    }
    protected function setAsTmStruct($value, $checkBoundaries = false) {
      if($value['tm_year'] === null) {
        $value['tm_year'] = date('Y') - 1900;
      }    
      $res = mktime(
        $value['tm_hour'],
        $value['tm_min'],
        $value['tm_sec'],
        $value['tm_mon']+1,
        $value['tm_mday'],
        $value['tm_year'] + 1900
      );
      if($res === false) {
        $this->setError(E_CONVERTOR_INVALID_DATETIME,array($value));      
        return false;
      }
      if($checkBoundaries) {
        if(($min = $this->getMinValue()) && ($res < $min)) {
          $this->setError(E_CONVERTOR_TOO_LOW,array($this->formatValue($this->inputFormat,$res),$this->formatValue($this->inputFormat,$min)));
          return false;
        }
        if(($max = $this->getMaxValue()) && ($res > $max)) {
          $this->setError(E_CONVERTOR_TOO_HIGH,array($this->formatValue($this->inputFormat,$res),$this->formatValue($this->inputFormat,$max)));
          return false;
        }
      }
      return $this->setRaw($res);
    }
    /**
    * Try to parse value with the given format
    *
    * @param string Value to be parsed
    * @param string Format to be used
    * @return bool Success
    */
    protected function parseValue($value, $format, $checkBoundaries = false) {
      $res = strptime($value,$format);
      if(($res === false) || (trim($res['unparsed']) != '')) {
        $this->setError(E_CONVERTOR_INVALID_DATETIME,array($value));
        return false;
      }
      $res = $this->setAsTmStruct($res,$checkBoundaries);
      if($res === false) {
        return false;
      }
      return true;       
    }
    /**
    * Try to parse value with the strtotime function
    *
    * @param string Value to be parsed
    * @return bool Success
    */    
    protected function parseValueStrtotime($value) {
      $res = strtotime($value);
      if(($res === false) || ($res === -1)) {
        $this->setError(E_CONVERTOR_INVALID_DATETIME,array($value));
        return false;
      }
      return $this->setRaw($res);
    }
    private function formatValue($format, $value = null) {
      if($value === null) {
        $value = $this->getRaw();
      }
      if($value === null) {
        return '';
      } else {
        return strftime($format,$value); 
      }
    }            
    public function getAsSql() {
      return strftime($this->sqlFormat,$this->getRaw());
    }
    public function setAsSQL($value) {
      if(($time = strtotime($value)) !== false) {
        $time = $this->setRaw($time);
      } 
      if($time === false) {
        $this->setError(E_CONVERTOR_INVALID_DATETIME,array($value));
        return false;
      }
      return true;
    }
    public function getAsInput() {
      return $this->formatValue($this->inputFormat);
    }
    public function setAsInput($value) {        
      return $this->parseValue($value,$this->inputFormat,true);    
    }
    public function getAsDisplay() {
      return $this->formatValue($this->displayFormat);
    }
    public function setAsDisplay($value) {
      return $this->parseValue($value,$this->displayFormat);
    }
    /**
    * Set the value to current time 
    *
    * @return void
    */
    public function setNow() {
      $this->setRaw(time());    
    }
    /**
    * Is the date is in daylight saving time? 
    *
    * @return bool
    */
    public function isDayLightSavingTime() {
      return (date('I',$this->getRaw()) == 1);      
    }
    /**
    * Get the timezone identifier 
    *
    * @return string
    */
    public function getTimeZone() {
      return date('e',$this->getRaw());      
    }
    /**
    * Get the timezone offset 
    *
    * @return string
    */
    public function getTimeZoneOffset() {
      return date('O',$this->getRaw());      
    }
    /**
    * Get date-time in RFC 2822 format
    *
    * @return string
    */
    public function getAsRfc2822()  {
      return date('r',$this->getRaw());
    }
    /**
    * Set date-time in RFC 2822 format
    *
    * @patam string
    * @return bool
    */
    public function setAsRfc2822($value) {
      return $this->parseValueStrtotime($value);
    }
    /**
    * Get date-time in ISO 8601 format
    *
    * @return string
    */
    public function getAsIso8601()  {
      return date('c',$this->getRaw());
    }    
    /**
    * Set date-time in ISO 8601 format
    *
    * @patam string
    * @return bool
    */
    public function setAsIso8601($value) {
      return $this->parseValueStrtotime($value);
    }    
    private function inc($value, $interval) {
      if(!is_numeric($value)) {
        $this->setError("Value to increment must be integer");
        return false;
      }
      $value += 0;
      if($value >= 0) {
        $value = '+'.$value;
      } 
      $val = strtotime($value.' '.$interval,$this->getRaw());
      if(($val === false) || ($val === -1)) {
        $this->setError("Couldn't increment ".$interval." by ".$value);
        return false;
      }
      return $this->setRaw($val);
    }
    /**
    * Add $value days
    *
    * @patam int number of days to add
    * @return bool
    */
    public function incDay($value = 1) {
      $this->inc($value,'day');    
    }
    /**
    * Add $value weeks
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incWeek($value = 1) {
      $this->inc($value,'week');    
    }    
    /**
    * Add $value hours
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incHour($value = 1) {
      $this->inc($value,'hour');    
    }
    /**
    * Add $value minutes
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incMinute($value = 1) {
      $this->inc($value,'minute');    
    }
    /**
    * Add $value seconds
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incSecond($value = 1) {
      $this->inc($value,'second');    
    }
    /**
    * Add $value month
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incMonth($value = 1) {
      $this->inc($value,'month');    
    }
    /**
    * Add $value year
    *
    * @patam int number of week to add
    * @return bool
    */
    public function incYear($value = 1) {
      $this->inc($value,'year');    
    }    
    private function setDayOfWeek($direction, $dayOfWeek) {
      $directions = array(
        1 =>'next',
        -1 => 'last'        
      );
      $daysOfWeek = array(
        0 => 'Sunday',
        1 => 'Monday',
        2 => 'Tuesday',
        3 => 'Wednesday',
        4 => 'Thursday',
        5 => 'Friday',
        6 => 'Saturday'
      );      
      if(!isset($directions[$direction]) || !isset($daysOfWeek[$dayOfWeek])) {
        $this->setError("dayOfWeek has to be an integer betwen 0 and 6");
        return false;        
      }
      $val = strtotime($directions[$direction].' '.$daysOfWeek[$dayOfWeek],$this->getRaw());
      if(($val === false) || ($val === -1)) {
        $this->setError("Couldn't set date to ".$interval." by ".$directions[$direction].' '.$daysOfWeek[$dayOfWeek]);
        return false;
      }
      return $this->setRaw($val);      
    }
    /**
    * Set value to next $dayOfWeek (0..6)
    *
    * @patam int Day of week (0..6)
    * @return bool
    */
    public function nextDayOfWeek($dayOfWeek) {
      $this->setDayOfWeek(1,$dayOfWeek);    
    }    
    /**
    * Set value to last $dayOfWeek (0..6)
    *
    * @patam int Day of week (0..6)
    * @return bool
    */
    public function lastDayOfWeek($dayOfWeek) {
      $this->setDayOfWeek(-1,$dayOfWeek);    
    }
  }
  class TDate extends TDateTime {
    public $displayFormat = LNG_DATE_FORMAT;
    public $inputFormat = LNG_DATE_FORMAT;
    public $sqlFormat = CFG_SQL_DATE_FORMAT;
  }
  class TTime extends TDateTime {
    public $displayFormat = LNG_TIME_FORMAT;
    public $inputFormat = LNG_TIME_FORMAT;
    public $sqlFormat = CFG_SQL_TIME_FORMAT;
  }
  class TLongTime extends TDateTime {
    public $displayFormat = LNG_LONG_TIME_FORMAT;
    public $inputFormat = LNG_LONG_TIME_FORMAT;
    public $sqlFormat = CFG_SQL_TIME_FORMAT;
  }
  class TShortDate extends TDateTime {
    public $displayFormat = LNG_SHORT_DATE_FORMAT;
    public $inputFormat = LNG_SHORT_DATE_FORMAT;
    public $sqlFormat = CFG_SQL_DATE_FORMAT;
  }
  class TShortDateTime extends TDateTime {
    public $displayFormat = LNG_SHORT_DATE_TIME_FORMAT;
    public $inputFormat = LNG_SHORT_DATE_TIME_FORMAT;
  }
  class TLongDateTime extends TDateTime {
    public $displayFormat = LNG_LONG_DATE_TIME_FORMAT;
    public $inputFormat = LNG_LONG_DATE_TIME_FORMAT;
  }
  class TNumberFormatOptions extends TObject {
    /** @var int Number of digits after the decimal point */
    public $decimals = -1;
    /** @var bool Force number of decimals */
    public $forceDecimals = false;
    /** @var string Thousand groups separator */
    public $thousandsSeparator = LNG_THOUSANDS_SEPARATOR;
    /** @var string Decimal point */
    public $decimalPoint = LNG_DECIMAL_POINT;
    /** @var bool Use scientific notation? (e.g. "1.2e+2")*/
    public $scientific = false;
  }
  class TNumber extends TCustomValue {
    /** @var TNumberFormatOptions Format options used for direct output */
    public $displayFormat;
    /** @var TNumberFormatOptions Format options used for input boxes */
    public $inputFormat;
    /** @var int Minimal allowed value */
    public $minValue = null;
    /** @var int Maximal allowed value */
    public $maxValue = null;
    
    public function __construct($parent = null) {
      parent::__construct($parent);
      $this->displayFormat = new TNumberFormatOptions();
      $this->inputFormat = new TNumberFormatOptions();
      $this->inputFormat->thousandsSeparator = LNG_INPUT_THOUSANDS_SEPARATOR;      
    }
    private function formatValue($value, TNumberFormatOptions $format) {
      if($value === null) {
        return '';
      }
      if($format->scientific) {
        return str_replace('.',$format->decimalPoint,sprintf('%e',$value));
      } elseif($format->forceDecimals && $format->decimals >= 0) {
          return number_format($value,
            $format->decimals,
            $format->decimalPoint,
            $format->thousandsSeparator); 
      } else {
        if($format->decimals == 0) {
          return number_format($value,
            0,
            $format->decimalPoint,
            $format->thousandsSeparator); 
        } elseif($format->decimals >= 0) {
          $val = round($value,$format->decimals);
          $val = number_format($val,
            $format->decimals,
            $format->decimalPoint,
            $format->thousandsSeparator);
          return rtrim(rtrim($val,'0'),$format->decimalPoint);                                
        } else {
          $dec = $value - round($value);
          $dec = strlen($dec)-2;
          if($dec < 0) {
            $dec = 0;
          }
          return number_format($value,
            $dec,
            $format->decimalPoint,
            $format->thousandsSeparator);
        }
      }
    }
    private function setValue($value,TNumberFormatOptions $format) {
      $val = str_replace(array(' ',$format->thousandsSeparator,$format->decimalPoint),array(' ','','.'),$value);
      if(!is_numeric($val)) {
        $this->setError(E_CONVERTOR_INVALID_NUMBER,array($value));
        return false;
      }
      $value = $val+0;
      if(($format->decimals == 0) && (($val - round($value)) != 0)) {
        $this->setError(E_CONVERTOR_INVALID_INTEGER,array($value));
        return false;
      }
      if($this->minValue !== null) {
        if($value < $this->minValue) {
          $this->setError(E_CONVERTOR_TOO_LOW,array($value,$this->minValue));
          return false;
        }
      }
      if($this->maxValue !== null) {
        if($value > $this->maxValue) {
          $this->setError(E_CONVERTOR_TOO_HIGH,array($value,$this->maxValue));
          return false;
        }
      }
      return $this->setRaw($value);
    }
    public function setAsSQL($value) {
      if($value === null) {
        return $this->setRaw(null);
      }
      if(!is_numeric($value)) {
        $this->setError(E_CONVERTOR_INVALID_NUMBER,array($value));
        return false;
      } 
      return $this->setRaw($value+0);
    }
    public function getAsInput() {
      return $this->formatValue($this->getRaw(),$this->inputFormat);
    }
    public function setAsInput($value) {
      return $this->setValue($value,$this->inputFormat);
    }
    public function getAsDisplay() {
      return $this->formatValue($this->getRaw(),$this->displayFormat);
    }
    public function getAsSql() {
      return str_replace(',','.',$this->getRaw());
    }    
    public function setAsDisplay($value) {
      return $this->setValue($value,$this->displayFormat);
    }
    public function getInputExample()  {
      return $this->formatValue(331203.389110398,$this->inputFormat);    
    }    
  }
  class TCurrency extends TNumber {
    /** @var string Currency string (if begins with "<" currency will be positioned before the number) */
    public $currency = LNG_CURRENCY;    
    public function __construct($parent = null) {
      parent::__construct($parent);
      $this->displayFormat->decimals = 2;
      $this->displayFormat->forceDecimals = true;            
    }
    public function getAsDisplay() {
      $val = parent::getAsDisplay();

      if($this->currency) {
        if (startsStr('<',$this->currency)) {
          $val = strDelete($this->currency,0,1).$val;
        } else {
          $val .= $this->currency;
        } 
      }
      return $val;
    }
  }
  /**
  *
  */
  class TText extends TCustomValue {
    /** @var int Maximal lenght of the text */
    public $maxLength = null;
    /** @var int Minimal lenght of the text */
    public $minLength = null;    
    /** @var bool Trim input text? */
    public $trimInput = false;    
    public function setAsInput($value) {
      $length = strLength($value);

      if(($this->maxLength !== null) && ($length > $this->maxLength)) {
        echo $length;
        $this->setError(E_CONVERTOR_TOO_LONG,array(strIntro($value,30),$this->maxLength));
        return false;
      }
      if(($this->minLength !== null) && ($length < $this->minLength)) {
        $this->setError(E_CONVERTOR_TOO_SHORT,array(strIntro($value,30),$this->minLength));
        return false;
      }
      if($this->trimInput) {
        $value = trim($value);
      }
      return $this->setRaw($value);
    }
  }
  /**
  *
  */
  class TValue extends TCustomValue {
  
  }

  /**
  *
  */
  class TBoolean extends TCustomValue {
    public function getInputExample()  {
      return LNG_YES;    
    }
    public function getAsSql() {
      return ($this->getRaw() ? '1' : '0');
    }
    public function setAsSQL($value) {
      if($value) {
        return $this->setRaw(true);
      } else {
        return $this->setRaw(false);
      }
      return true;
    }
    public function getAsInput() {
      return $this->getAsDisplay();
    }
    public function setAsInput($value) {
      return $this->setAsDisplay($value);    
    }
    public function getAsDisplay() {
      return ($this->getRaw() ? LNG_YES : LNG_NO);
    }
    public function setAsDisplay($value) {
      return $this->setRaw(sameText($value,LNG_YES));
    }
  } 
?>